#include <QApplication>
#include <QThread>
#include <QProgressDialog>
#include <QGraphicsEllipseItem>
#include "analyzer.h"
#include "ui_analyzer.h"
#include "datastream.h"
#include "datalog.h"
#include "config.h"
#include "common.h"

analyzer_ui::analyzer_ui(datalog *log_in, QWidget *parent) :
  QWidget(parent),
  ui(new Ui::analyzer) {
  ui->setupUi(this);
  // -------------
  anl = new analysis; // new analysis object
  log = log_in;
}

analyzer_ui::~analyzer_ui() {
  delete anl;
  delete ui;
}

void analyzer_ui::on_interface_loaded() {
  anl->min_afrcounts = ui->anl_afrcounts_slider->value();
  anl->min_cooltmp = ui->anl_mincooltmp_slider->value();
  anl->min_timestamp = ui->anl_mintime_slider->value();
  anl->use_integrator = ui->setting_anl_int->isChecked();
  anl->anl_filter_idleanddecel = ui->setting_anl_extracells->isChecked();
  anl->wbpercent = ui->setting_display_wb_percent->isChecked();
  anl->worst_case_blm = ui->setting_afrblmavg->isChecked();
  ui->anl_knockchart->setScene(anl->knock_display);
  wideband_config_updated(); // enables/disables wideband
  if(config.get_analyze_wideband() == true) { // remember wb/nb setting
    ui->afr_input_wide->setChecked(true);
  } else {
    ui->afr_input_narrow->setChecked(true);
  }
  label_anl_graphs();
}

//--------------------------------------------

void analyzer_ui::on_analyzeDataBtn_clicked() {
  // routine to check if any data exists ...
  if(ui->anl_all_btn->isChecked() == false && log->is_data() == false) { // if only one log and it's empty
    emit info_dialog("No recorded data to analyze.\nEither record some data, or use the load button on the datalogger page and select 'analyze all logs'.");
    return;
  } else if(ui->anl_all_btn->isChecked() == true) { // if multiple logs
    // search for any data
    bool any_data = false;
    datalog *search_log = log->get_first();
    while(search_log != nullptr) {
      if(search_log->is_data() == true) {
        any_data = true;
        break;
      }
      search_log = search_log->get_next();
    }
    if(any_data == false) {
      emit info_dialog("No data to analyze in any loaded log.\nUse the LOAD button on the datalogger page to load log data.");
      return;
    }
  }
  //-------------
  anl->reset_analyzer();
  ui->analyzeDataBtn->setEnabled(false);
  //-------------
  datalog *anl_log;
  if(ui->anl_all_btn->isChecked() == true) {
    anl_log = log->get_first();
  } else {
    anl_log = log;
  }
  //-------------
  // delete lock all logs
  datalog *lock_log = log->get_first();
  while(lock_log != nullptr) {
    lock_log->del_lock.lock();
    lock_log = lock_log->get_next();
  }
  //-------------
  QProgressDialog progress("Analyzing data...","ABORT",0,1);
  progress.setMinimumDuration(0);
  progress.setWindowModality(Qt::WindowModal);
  progress.setAutoClose(false);
  //-------------
  while(anl_log != nullptr) {
    // analyze
    if(anl_log->is_data() == true) {
      progress.setRange(0,anl_log->get_pkt_count());
      datalog_packet *x = anl_log->oldest_packet();
      while(x->is_last_packet() == false) {
        anl->analyze_segment(x);
        x = x->get_next();
        progress.setValue(x->sequence);
        if(progress.wasCanceled()) {
          goto ANL_CANCEL;
        }
      }
    }
    // advance
    if(ui->anl_all_btn->isChecked() == true) {
      anl_log = anl_log->get_next();
    } else {
      anl_log = nullptr; // breaks loop
    }
  }
  //-------------
  ANL_CANCEL:
  //-------------
  draw_analysis_results(); // set fields
  progress.cancel();
  //-------------
  ui->analyzeDataBtn->setEnabled(true);
  //-------------
  // clear delete locks
  // delete lock all logs
  lock_log = log->get_first();
  while(lock_log != nullptr) {
    lock_log->del_lock.unlock();
    lock_log = lock_log->get_next();
  }
  //-------------
  QApplication::restoreOverrideCursor();
}

void analyzer_ui::wideband_config_updated() {
  if(config.get_wideband_enable() == true) {
    ui->afr_input_wide->setEnabled(true);
  } else {
    ui->afr_input_wide->setEnabled(false);
    ui->afr_input_narrow->setChecked(true);
    config.set_analyze_wideband(false);
  }
}

void analyzer_ui::def_parse_err_in(QString str) {
  emit error_dialog(str);
}

void analyzer_ui::on_anl_mincooltmp_slider_valueChanged(int value) {
   ui->anl_mincooltmp_display->setPlainText(QString("%1").arg(value));
   anl->min_cooltmp = value;
}

void analyzer_ui::on_anl_mintime_slider_valueChanged(int value) {
  ui->anl_mintime_display->setPlainText(QString("%1").arg(value));
  anl->min_timestamp = value;
}

void analyzer_ui::on_anl_afrcounts_slider_valueChanged(int value) {
  ui->anl_afrcounts_display->setPlainText(QString("%1").arg(value));
  anl->min_afrcounts = value;
}

void analyzer_ui::on_afr_input_narrow_toggled(bool checked) {
  if(checked == false) return;
  ui->grp_nbfilter->setEnabled(true);
  ui->group_wbfilter->setEnabled(false);
  anl->wideband = !checked;
  ui->maf_trim_label->setText("TRIM");
}

void analyzer_ui::on_afr_input_wide_toggled(bool checked) {
  if(checked == false) return;
  ui->grp_nbfilter->setEnabled(false);
  ui->group_wbfilter->setEnabled(true);
  anl->wideband = checked;
  ui->maf_trim_label->setText("AFR");
  config.set_analyze_wideband(true);
}

//--------------------------------------------

fcell::fcell() {
  low = 0;
  high = 0;
  average_accumulator = 0;
  count = 0;
}

void fcell::clear() {
  low = 0;
  high = 0;
  average_accumulator = 0;
  count = 0;
}

float fcell::avg() {
  if(count > 1) {
    return average_accumulator / count;
  } else if(count < 0) {
    return 0;
  } else {
    return average_accumulator;
  }
}

void fcell::add_data(float value) {
  // analyze
  if(count == 0) { // first record
    high = value;
    low = value;
  } else {
    if(value > high) high = value;
    if(value < low) low = value;
  }
  average_accumulator += value;
  count += 1;
}

void fcell::add_data(float value, float min, float max) {
  if(value > min && value < max) add_data(value);
}

//--------------------------------------------

analysis::analysis() {
  min_cooltmp = 50;
  min_timestamp = 60;
  knock_display = new QGraphicsScene;
  worst_case_blm = false;
  wideband = false;
  wbcl = false;
  wbpercent = false;
  reset_analyzer();
}

void analyzer_ui::on_setting_anl_int_toggled(bool checked) {
  anl->use_integrator = checked;
}

void analyzer_ui::on_setting_anl_extracells_toggled(bool checked) {
  anl->anl_filter_idleanddecel = checked;
}

void analyzer_ui::on_setting_afrblmavg_toggled(bool checked) {
  anl->worst_case_blm = checked;
}

void analyzer_ui::on_setting_wb_cl_toggled(bool checked) {
  anl->wbcl = checked;
}

void analyzer_ui::on_setting_enh_accuracy_toggled(bool checked) {
   anl->enhanced_accuracy = checked;
}

void analyzer_ui::on_setting_display_wb_percent_toggled(bool checked) {
   anl->wbpercent = checked;
}

void analysis::reset_analyzer() {
  lo2.clear();
  ro2.clear();
  idle_trim[0].clear();
  idle_trim[1].clear();
  idle_map.clear();
  idle_maf.clear();
  lcrosscount = 0; rcrosscount = 0;
  lstate = 0; rstate = 0;
  rstuck = 0; lstuck = 0;
  total_knock_events = 0;
  total_knock_counts = 0;
  knock_display->clear();
  config_knock_graph();
  for(int x=0;x<14;x++) {
    for(int y=0;y<10;y++) {
      ve_trim[x][y].clear();
    }
  }
  for(int x=0;x<70;x++) {
    maf_trim[x].clear();
  }
  for(int rpm_axis=0;rpm_axis<28;rpm_axis++) {
    for(int x=0;x<3;x++) pe_fuel[rpm_axis][x].clear();
  }
  for(int x=0;x<19;x++) {
    for(int y=0;y<4;y++) {
      blm[x][y].clear();
    }
  }
}

void analysis::analyze_segment(datalog_packet *p) {
  if(p->get_msgnumber() != 0x00 ||
     (p->get_msgnumber() == 0x05 &&
      p->log->info.patch_version.get() >= 0x01)) return;
  parse_record = p;
  // bail if out of global thresholds
  if(p->get_float("COOLTMP") < min_cooltmp) return;
  if(parse_record->timestamp < min_timestamp * 1000) return;
  if(p->get_float("RPM") < 200) return; // eng probably not running

  // run various analyzers
  analyze_o2();
  if(wideband == true) {
    analyze_wb_afr();
  } else {
    analyze_nb_afr();
  }
  analyze_idle();
  analyze_pe();
  analyze_knock();
  analyze_blm();
}

//--------------------------------------------

void analyzer_ui::draw_results_o2() {
  //------LH
  ui->anl_o2record_l->setPlainText(QString("%1").arg(anl->lo2.count));
  ui->anl_peakmv_l->setPlainText(QString("%1mv").arg(anl->lo2.high));
  ui->anl_lowmv_l->setPlainText(QString("%1mv").arg(anl->lo2.low));
  ui->anl_avgmv_l->setPlainText(QString("%1mv").arg(anl->lo2.avg()));
  ui->anl_crosscount_l->setPlainText(QString("%1").arg(anl->lcrosscount));
  //------RH
  ui->anl_peakmv_r->setPlainText(QString("%1mv").arg(anl->ro2.high));
  ui->anl_lowmv_r->setPlainText(QString("%1mv").arg(anl->ro2.low));
  ui->anl_avgmv_r->setPlainText(QString("%1mv").arg(anl->ro2.avg()));
  ui->anl_crosscount_r->setPlainText(QString("%1").arg(anl->rcrosscount));
}

//--------------------------------------------

void analysis::analyze_pe() {
  // filtering
  if(parse_record->get_bool("PE") == false) return;

  // get vars
  int current_rpm = parse_record->get_int("RPM");
  float current_map = parse_record->get_float("MAP");
  int rpm_axis = select_cell(7000,28,current_rpm);
  int map_cell = 0;
  if(current_map < 70) {
    map_cell = 0;
  } else if(current_map > 90) {
    map_cell = 2;
  } else { // (70-90)
    map_cell = 1;
  }

  float current_trim;
  if(config.get_wideband_enable() == false) { // regardless of nb/wb setting, use wideband in pe
    current_trim = ( parse_record->get_float("LO2") + parse_record->get_float("RO2") ) / 2;
  } else {
    current_trim = parse_record->get_float(config.get_wb_str());
  }

  pe_fuel[rpm_axis][map_cell].add_data(current_trim);
}

//--------------------------------------------

void analysis::analyze_idle() {
  if(parse_record->get_int("BLMCELL") != 16) return;
  if(wideband == true) {
    if(wbcl != parse_record->get_bool("CL")) return;
    idle_trim[0].add_data(parse_record->get_float(config.get_wb_str()));
    idle_trim[1].add_data(parse_record->get_float(config.get_wb_str()));
    idle_map.add_data(parse_record->get_float("MAP"));
    idle_maf.add_data(parse_record->get_float("MAF"),1,99999);
  } else {
    if(parse_record->get_bool("CL") == false) return;
    idle_trim[0].add_data(parse_record->get_int("LBLM"));
    idle_trim[1].add_data(parse_record->get_int("RBLM"));
    idle_map.add_data(parse_record->get_float("MAP"));
    idle_maf.add_data(parse_record->get_float("MAF"),1,99999);
  }

}

//--------------------------------------------

void analysis::analyze_blm() {
  if(parse_record->get_bool("CL") == false) return;
  if(parse_record->get_bool("PE") == true) return;
  int blmcell = clamp(parse_record->get_int("BLMCELL"),0,18);
  blm[blmcell][0].add_data(parse_record->get_float("RPM"));
  blm[blmcell][1].add_data(parse_record->get_float("MAP"));
  blm[blmcell][2].add_data(parse_record->get_int("LBLM"));
  blm[blmcell][3].add_data(parse_record->get_int("RBLM"));
}

//--------------------------------------------

void analysis::analyze_nb_afr() {
  //filtering
  if(parse_record->get_bool("CL") == false) return;
  if(parse_record->get_bool("PE") == true) return;
  if(parse_record->get_bool("BLM") == false) return;
  if(parse_record->get_int("BLMCELL") == 18) return;
  if(anl_filter_idleanddecel == true) {
    if(parse_record->get_int("BLMCELL") >= 16) return;
    if(parse_record->get_float("TPS") <= 1) return; // no tps
    if(parse_record->get_float("LBPW") <= 1) return; // no lbpw
    if(parse_record->get_float("RBPW") <= 1) return; // no rbpw
  }
  // get vars
  float current_blm;
  int lblm = parse_record->get_int("LBLM");
  int rblm = parse_record->get_int("RBLM");
  if(worst_case_blm == true) {
    if(lblm > rblm) { // left higher
      current_blm = lblm;
    } else { // right higher
      current_blm = rblm;
    }
  } else {
    current_blm = clamp(( lblm + rblm ) / 2,0,255);
  }
  float current_int = clamp(( parse_record->get_int("LINT") + parse_record->get_int("RINT") ) / 2,0,255);
  if(enhanced_accuracy == true) {
    if(current_int < 112 || current_int > 140) return;
  }
  float trim = current_blm;
  if(use_integrator == true) {
    trim += ( current_int - 128 );
    trim = clamp((float)trim,(float)0,(float)255);
  }
  int current_rpm = parse_record->get_int("RPM");
  int current_map = parse_record->get_int("MAP");
  // ve_trim[14][10]
  int rpm_axis = select_cell(7000,14,current_rpm);
  int map_axis = select_cell(100,10,current_map);
  ve_trim[rpm_axis][map_axis].add_data(trim);
  // maf
  int current_afgs = parse_record->get_int("MAF");
  if(current_afgs < 1) return; // no airflow
  int afgs_axis = select_cell(300,70,current_afgs);
  maf_trim[afgs_axis].add_data(trim);
}

void analysis::analyze_wb_afr() {
  //filtering
  if(parse_record->get_bool("PE") == true) return;
  if(wbcl == true) {
    if(parse_record->get_bool("CL") == false) return;
    if(parse_record->get_bool("BLM") == false) return;
  } else {
    if(parse_record->get_bool("CL") == true) return;
    if(parse_record->get_bool("BLM") == true) return;
  }
  if(parse_record->get_float("LBPW") < 0.01 || parse_record->get_float("RBPW") < 0.01) return; // dfco
  // get vars
  float current_afr = parse_record->get_float(config.get_wb_str());
  float current_target;
  current_target = parse_record->get_float("AFR");
  if(wbpercent == true) {
    if(current_target < 1) return; // no afr target data in percentage mode, err!
    current_afr = current_afr / current_target;
  }
  int current_rpm = parse_record->get_int("RPM");
  int current_map = parse_record->get_int("MAP");
  // ve_trim[14][10]
  int rpm_axis = select_cell(7000,14,current_rpm);
  int map_axis = select_cell(100,10,current_map);
  ve_trim[rpm_axis][map_axis].add_data(current_afr);
  // maf
  int current_afgs = parse_record->get_int("MAF");
  if(current_afgs < 1) return; // no airflow
  int afgs_axis = select_cell(300,70,current_afgs);
  maf_trim[afgs_axis].add_data(current_afr);
}

//--------------------------------------------

int analysis::select_cell(int range, int cells, float data) {
  float interval = range / cells;
  int cell = floor((data/(float)range)*(float(range)/interval));
  if(cell > cells - 1) return cells - 1;
  if(cell < 0) return 0;
  return cell;
}

//--------------------------------------------

void analysis::add_knock_point(int rpm, int map, int count, bool wot, int spark, int kr, int interval, timestamp_t timestamp) {
  int counts_per_msec = count/interval;
  int size = clamp(counts_per_msec,6,20); // get circle size
  float scale_x = KNOCK_TABLE_W * ( clamp((float)map,(float)5,(float)98) / 100);
  float scale_y = KNOCK_TABLE_H * ( (float)rpm / 7000);
  QGraphicsEllipseItem *x = knock_display->addEllipse(scale_x - (size / 2),
                            scale_y - (size / 2),
                            size, size);
  if(wot == true) { // color for wot
    x->setBrush(QColor(200,0,0,225));
  } else {
    x->setBrush(QColor(0,180,0,225));
  }
  x->setToolTip(QString("Knock Event Details\nIncrement: %1/msec\nRPM: %2\nMAP: %3 KPA\nSPARK: %4deg\nKR: %5\nTIMESTAMP: %6").
                arg(counts_per_msec).arg(rpm).arg(map).arg(spark).arg(kr).arg((float)timestamp / 1000));
}

void analysis::analyze_knock() {
  datalog_packet *prev = parse_record->get_prev(parse_record->get_device(),parse_record->get_msgnumber());
  if(prev == nullptr) return; // no prev data
  int prev_knock = prev->get_int("KCOUNT");
  timestamp_t prev_timestamp = prev->timestamp;
  int current_knock = parse_record->get_int("KCOUNT");
  timestamp_t current_timestamp = parse_record->timestamp;
  if(current_knock > prev_knock &&
     current_timestamp - prev_timestamp < 500) { // reject large rec gap
    if(total_knock_events >= 800) return; // bail if tons of records
    add_knock_point(parse_record->get_int("RPM"),parse_record->get_int("MAP"),
                    current_knock - prev_knock,
                    parse_record->get_bool("PE"),
                    parse_record->get_int("SPKADV"),
                    parse_record->get_float("KR"),
                    current_timestamp - prev_timestamp,
                    current_timestamp);
    total_knock_events++;
    total_knock_counts += ( current_knock - prev_knock );
  }
}

void analysis::config_knock_graph() {
  for(int x=1;x<=9;x++) { // draw vertical lines
    knock_display->addLine(((float)KNOCK_TABLE_W/10)*(float)x,1,
                           ((float)KNOCK_TABLE_W/10)*(float)x,KNOCK_TABLE_H - 1);
  }
  for(int x=1;x<=14;x++) { // hz lines
    knock_display->addLine(1,((float)KNOCK_TABLE_H/14)*(float)x,
                           KNOCK_TABLE_W - 1,((float)KNOCK_TABLE_H/14)*(float)x);
  }
}

//--------------------------------------------

void analysis::analyze_o2() {
  // filtering
  if(parse_record->get_bool("CL") == false) return;
  if(parse_record->get_bool("PE") == true) return;
  if(parse_record->get_bool("BLM") == false) return;

  int lo2v = parse_record->get_int("LO2");
  int ro2v = parse_record->get_int("RO2");
  lo2.add_data(lo2v);
  ro2.add_data(ro2v);
  // crosscounts
  if(lo2v > 600) { // left o2 high
    if(lstate == 0) { // state was low
      lstate = 1; // set state high
      lcrosscount++; // increment cross count
    }
  } else if(lo2v < 400) { // left o2 low
    if(lstate == 1) { // state was high
      lstate = 0; // set state low
      lcrosscount++;
    }
  } else {
    lstuck++;
  }
  if(ro2v > 600) { // r o2 high
    if(rstate == 0) { // state was low
      rstate = 1; // set state high
      rcrosscount++; // increment cross count
    }
  } else if(ro2v < 400) { // r o2 low
    if(rstate == 1) { // state was high
      rstate = 0; // set state low
      rcrosscount++;
    }
  } else {
    rstuck++;
  }
}

//--------------------------------------------

void analyzer_ui::label_anl_graphs() {
  for(int x=0;x<70;x++) {
    ui->afr_maf_table->setItem(x,0,
          new QTableWidgetItem(QString("%2-%1").arg((x + 1) * 5).arg(((x + 1) * 5)-4)));
  }
  for(int x=0;x<28;x++) {
    ui->afr_pe_table->setItem(x,0,
          new QTableWidgetItem(QString("%2-%1").arg((x + 1) * 250).arg(((x + 1) * 250)-250)));
  }
}

void analyzer_ui::draw_analysis_results() {
  draw_results_o2();
  // knock is drawn live already ...
  ui->total_knock_counts->setPlainText(QString("%1").arg(anl->total_knock_counts));
  ui->total_knock_events->setPlainText(QString("%1").arg(anl->total_knock_events));
  if(anl->wideband == true) {
    draw_results_wb_afr();
  } else {
    draw_results_nb_afr();
  }
  draw_results_pe();
  draw_results_idle();
  draw_results_blm();
}

QString analyzer_ui::percent_to_str(double x) {
  if(config.get_low_precision() == true) {
    if(x < DISCARD_TRIM && x > -DISCARD_TRIM && config.get_low_percent_ok() == true) {
      return QStringLiteral(" 0%");
    }
    if(x > 0) return QString("+%1%").arg((double)x,0,'f',0);
    if(x < 0) return QString("%1%").arg((double)x,0,'f',0); // already has minus sign ...
    return QStringLiteral(" 0%");
  } else { // one decimal place
    if(x < DISCARD_TRIM && x > -DISCARD_TRIM && config.get_low_percent_ok() == true) {
      return QStringLiteral(" 0.0");
    }
    if(x > 0) return QString("+%1").arg((double)x,0,'f',1);
    if(x < 0) return QString("%1").arg((double)x,0,'f',1); // already has minus sign ...
    return QStringLiteral(" 0.0");
  }
}

float analyzer_ui::trim_to_percent(int trim) {
  return (((float)trim / 128.00 * 100.00)-100.00);
}

void analyzer_ui::draw_results_idle() {
  if(anl->idle_trim[0].count > anl->min_afrcounts) {
    if(ui->setting_display_blm_percent->isChecked() == true && anl->wideband == false) {
      ui->anl_idle_lh->setPlainText(percent_to_str(trim_to_percent(anl->idle_trim[0].avg())));
    } else {
      ui->anl_idle_lh->setPlainText(QString::number(anl->idle_trim[0].avg(),'f',1));
    }
  } else {
    ui->anl_idle_lh->setPlainText("---");
  }
  if(anl->idle_trim[1].count > anl->min_afrcounts) {
    if(ui->setting_display_blm_percent->isChecked() == true && anl->wideband == false) {
      ui->anl_idle_rh->setPlainText(percent_to_str(trim_to_percent(anl->idle_trim[1].avg())));
    } else {
      ui->anl_idle_rh->setPlainText(QString::number(anl->idle_trim[1].avg(),'f',1));
    }
  } else {
    ui->anl_idle_rh->setPlainText("---");
  }
  if(anl->idle_map.count > anl->min_afrcounts) {
    ui->anl_idle_map->setPlainText(QString::number(anl->idle_map.avg(),'f',1));
  } else {
    ui->anl_idle_map->setPlainText("---");
  }
  if(anl->idle_maf.count > anl->min_afrcounts) {
    ui->anl_idle_maf->setPlainText(QString::number(anl->idle_maf.avg(),'f',1));
  } else {
    ui->anl_idle_maf->setPlainText("---");
  }
}

void analyzer_ui::draw_results_nb_afr() {
  bool display_percentage = ui->setting_display_blm_percent->isChecked();
  for(int map_axis=0;map_axis<10;map_axis++) {
    for(int rpm_axis=0;rpm_axis<14;rpm_axis++) {
      float trim = anl->ve_trim[rpm_axis][map_axis].avg();
      float trim_percent = (((double)trim / 128.0 * 100.0)-100.0);
      if(anl->ve_trim[rpm_axis][map_axis].count >= anl->min_afrcounts) {
        if(display_percentage == true) {
          ui->afr_ve_table->setItem(rpm_axis,map_axis, new QTableWidgetItem(percent_to_str(trim_percent)));
        } else {
          ui->afr_ve_table->setItem(rpm_axis,map_axis, new QTableWidgetItem(QString::number(trim,'f',0)));
        }
        if(trim > 134) ui->afr_ve_table->item(rpm_axis,map_axis)->setBackgroundColor(Qt::green);
        if(trim < 122)  ui->afr_ve_table->item(rpm_axis,map_axis)->setBackgroundColor(Qt::red);
      } else {
        ui->afr_ve_table->setItem(rpm_axis,map_axis, new QTableWidgetItem(QString(" ")));
        ui->afr_ve_table->item(rpm_axis,map_axis)->setBackgroundColor(Qt::darkGray);
      }
    }
  }
  for(int maf_axis=0;maf_axis<70;maf_axis++) {
    float trim = anl->maf_trim[maf_axis].avg();
    float trim_percent = (((double)trim / 128.0 * 100.0)-100.0);
    if(anl->maf_trim[maf_axis].count >= anl->min_afrcounts) {
      if(display_percentage == true) {
        ui->afr_maf_table->setItem(maf_axis,1, new QTableWidgetItem(percent_to_str(trim_percent)));
      } else {
        ui->afr_maf_table->setItem(maf_axis,1,new QTableWidgetItem(QString::number(trim,'f',0)));
      }
      if(trim > 134) ui->afr_maf_table->item(maf_axis,1)->setBackgroundColor(Qt::green);
      if(trim < 122) ui->afr_maf_table->item(maf_axis,1)->setBackgroundColor(Qt::red);
    } else {
      ui->afr_maf_table->setItem(maf_axis,1, new QTableWidgetItem(QString(" ")));
      ui->afr_maf_table->item(maf_axis,1)->setBackgroundColor(Qt::darkGray);
    }

  }
}

void analyzer_ui::draw_results_wb_afr() {
  for(int map_axis=0;map_axis<10;map_axis++) {
    for(int rpm_axis=0;rpm_axis<14;rpm_axis++) {
      float afr = anl->ve_trim[rpm_axis][map_axis].avg();
      float percent = clamp((float)((anl->ve_trim[rpm_axis][map_axis].avg() * 100.00) - 100.00),-99.00,99.00);
      // draw VE
      if(anl->ve_trim[rpm_axis][map_axis].count >= anl->min_afrcounts) {
        if(anl->wbpercent == false) {
          ui->afr_ve_table->setItem(rpm_axis,map_axis, new QTableWidgetItem(QString::number((float)afr,'f',1)));
          if(afr > 15) {
            ui->afr_ve_table->item(rpm_axis,map_axis)->setBackgroundColor(Qt::green);
          } else if(afr < 12) {
            ui->afr_ve_table->item(rpm_axis,map_axis)->setBackgroundColor(Qt::red);
          }
        } else { // percentage mode
          ui->afr_ve_table->setItem(rpm_axis,map_axis, new QTableWidgetItem(percent_to_str(percent)));
          if(percent < -5) ui->afr_ve_table->item(rpm_axis,map_axis)->setBackgroundColor(Qt::red);
          if(percent > 5) ui->afr_ve_table->item(rpm_axis,map_axis)->setBackgroundColor(Qt::green);
        }
      } else {
        ui->afr_ve_table->setItem(rpm_axis,map_axis, new QTableWidgetItem(QString(" ")));
        ui->afr_ve_table->item(rpm_axis,map_axis)->setBackgroundColor(Qt::darkGray);
      }
    }
  }
  // draw MAF
  for(int maf_axis=0;maf_axis<70;maf_axis++) {
    float afr = anl->maf_trim[maf_axis].avg();
    float percent = clamp((float)((afr * 100.00) - 100.00),-99.00,99.00);
    if(anl->maf_trim[maf_axis].count >= anl->min_afrcounts) {
      if(anl->wbpercent == false) { // raw mode
        ui->afr_maf_table->setItem(maf_axis,1,new QTableWidgetItem(QString::number((float)afr,'f',1)));
        if(afr > 15) {
          ui->afr_maf_table->item(maf_axis,1)->setBackgroundColor(Qt::green);
        } else if(afr < 12) {
          ui->afr_maf_table->item(maf_axis,1)->setBackgroundColor(Qt::red);
        }
      } else { // percent mode
        ui->afr_maf_table->setItem(maf_axis,1, new QTableWidgetItem(percent_to_str(percent)));
        if(afr < -5) ui->afr_maf_table->item(maf_axis,1)->setBackgroundColor(Qt::red);
        if(afr > 5) ui->afr_maf_table->item(maf_axis,1)->setBackgroundColor(Qt::green);
      }
    } else {
      ui->afr_maf_table->setItem(maf_axis,1, new QTableWidgetItem(QString(" ")));
      ui->afr_maf_table->item(maf_axis,1)->setBackgroundColor(Qt::darkGray);
    }
  }
}

void analyzer_ui::draw_results_blm() {
  for(int x=0;x<19;x++) {
    // cell number and count
    ui->anl_blm_chart->setItem(x,0,
                               new QTableWidgetItem(QString("%1 (%2)").
                                                    arg(x).
                                                    arg(anl->blm[x][0].count)));
    for(int y=0;y<2;y++) { //rpm/map
      if(anl->blm[x][0].count < 1) {
        ui->anl_blm_chart->setItem(x,y + 1,new QTableWidgetItem(QString(" ")));
      } else {
       ui->anl_blm_chart->setItem(x,y + 1,
                                 new QTableWidgetItem(
                                   QString("%1 - %2").
                                   arg(anl->blm[x][y].low,0,'f',0).
                                   arg(anl->blm[x][y].high,0,'f',0)));
      }
    }
    // blm averages
    if(anl->blm[x][0].count < 1) { // not enough data
      ui->anl_blm_chart->setItem(x,3,new QTableWidgetItem(QString(" ")));
    } else {
      // lblm
      ui->anl_blm_chart->setItem(x,3,
                                new QTableWidgetItem(QString("%1").arg(anl->blm[x][2].avg(),0,'f',0)));
      // rblm
      ui->anl_blm_chart->setItem(x,4,
                                new QTableWidgetItem(QString("%1").arg(anl->blm[x][3].avg(),0,'f',0)));
    }
  }
}

void analyzer_ui::draw_results_pe() {
  for(int rpm_axis=0;rpm_axis<28;rpm_axis++) {
    for(int map_cell=0;map_cell<3;map_cell++) {
      float wbfuel = anl->pe_fuel[rpm_axis][map_cell].avg();
      if(config.get_wideband_enable() == false) { // use narrowband mv
        if(anl->pe_fuel[rpm_axis][map_cell].count < 3) { // not enough data
          ui->afr_pe_table->setItem(rpm_axis,map_cell + 1,new QTableWidgetItem(QString(" ")));
          ui->afr_pe_table->item(rpm_axis,map_cell + 1)->setBackgroundColor(Qt::darkGray);
          continue;
        } // else valid data
        ui->afr_pe_table->setItem(rpm_axis,map_cell + 1,
                                  new QTableWidgetItem(QString::number(wbfuel,'f',0)));
        if(wbfuel > 825) {
          ui->afr_pe_table->item(rpm_axis,map_cell + 1)->setBackgroundColor(Qt::green);
        } else if(wbfuel <= 825) {
          ui->afr_pe_table->item(rpm_axis,map_cell + 1)->setBackgroundColor(Qt::red);
        }
      } else { // wideband afr
        if(anl->pe_fuel[rpm_axis][map_cell].count < 4) { // not enough data
          ui->afr_pe_table->setItem(rpm_axis,map_cell + 1,new QTableWidgetItem(QString(" ")));
          ui->afr_pe_table->item(rpm_axis,map_cell + 1)->setBackgroundColor(Qt::darkGray);
          continue;
        } // else valid data
        ui->afr_pe_table->setItem(rpm_axis,map_cell + 1,new QTableWidgetItem(QString::number(wbfuel,'f',1) + ":1"));
          if(wbfuel > 14) ui->afr_pe_table->item(rpm_axis,map_cell + 1)->setBackgroundColor(Qt::green);
          if(wbfuel < 12) ui->afr_pe_table->item(rpm_axis,map_cell + 1)->setBackgroundColor(Qt::red);
      }
    }
  }
}


