#include <QApplication>
#include <QThread>
#include <QtDebug>
#include <chrono>
#include <thread>
#include "datastream.h"
#include "datalog.h"
#include "config.h"
#include "flashmagic.h"
#include "common.h"
#include <QDebug>

bool datastream::read_flash_section(byte side, byte *bin, quint16 offset) {
 flash_err.reset();
 // construct request
 buf_out[BDEVICE] = side;
 buf_out[BMODE] = 0x06; /* mode request */
 buf_out[BMSG] = 0x02; /* message */
 buf_out[4] = 0x00;
 buf_out[5] = 0x20;
 buf_out[6] = 0x0b;
 buf_out[7] = (offset >> 8) & 0x00FF; /* memory location high byte */
 buf_out[8] = offset & 0x00FF; /* memory location low byte */
 // attempt counter
 int attempts = 1;
 RETRY_FLASH_READ:
 // send request
 if(aldl_request_reply_length(10,135,250,FLASH_WRITE_POST_DELAY_SHORT) == false) {
   attempts++;
   if(attempts < 4) goto RETRY_FLASH_READ;
   debug.print("FLASH: REQUEST FAIL FOR FLASH PKT");
   flash_err.set_errstring("Can't read ECM data, giving up!");
   return false;
 }
 // verify confirmation byte
 if(buf_in[3] != 0xAA) {
   attempts++;
   if(attempts < 4) goto RETRY_FLASH_READ;
   debug.print("FLASH: NO CONFIRMATION BYTE IN FLASH PKT");
   flash_err.set_errstring("Can't read ECM data, giving up!");
   return false;
 }
 // copy data (offset in datastream is 6)
 for(int x=0;x<128;x++) {
   bin[x + offset] = buf_in[x + 6];
 }
 return true;
}

bool datastream::enter_mode5(byte side) {
  flash_err.reset();
  buf_out[BDEVICE] = side;
  buf_out[BMODE] = 0x05;
  // expected reply.  if mismatch, probably security violation.
  byte reply[5];
  reply[BMSG] = 0xAA;
  if(aldl_request_reply_specific(4,reply,5) == false) {
    debug.print("FLASH: Mode 5 Entry Fail");
    flash_err.set_errstring("Can't enter program upload mode (Mode 5), code: ",buf_in[BMSG]);
    return false;
  }
  return true;
}

void datastream::tickle_ecm() {
  for(int x=1;x<=4;x++) {
    sleep_ms(20);
    send_keepalive();
  }
}

bool datastream::unlock_security(byte side) {
  flash_err.reset();
  // attempt counter
  int attempts = 0;
  // generate mode 13 requests
  byte reply[7];
  // following are static throughout operation, ecm# and mode
  buf_out[BDEVICE] = side;
  buf_out[BMODE] = 0x0d; // mode
  // --- GENERATE REQUEST FOR SEED
  buf_out[BMSG] = 0x01; // req seed value msg
  // SEND REQUEST FOR KEY
  attempts = 0;
  RETRY_KEY_SEED_REQUEST:
  if(aldl_request_reply_length(5,7,200,75) == false) {
    if(attempts >= 4) {
      flash_err.set_errstring("Can't get the security key seed (Mode 13), giving up.");
      return false; // UP TO 2 TRIES
    }
    attempts++;
    tickle_ecm(); // poke the ecm a bit and maybe it'll generate a seed
    goto RETRY_KEY_SEED_REQUEST;
  }
  // VALIDATE
  if(buf_in[3] != 0x01) {
    debug.print("FLASH: Incorrect seed request response.");
    if(attempts >= 4) {
      flash_err.set_errstring("The ECM isn't responding properly to our security seed request.");
      return false; // UP TO 2 TRIES
    }
    attempts++;
    tickle_ecm(); // poke the ecm a bit and maybe it'll generate a seed
    goto RETRY_KEY_SEED_REQUEST;
  }
  // possible responses
  if(buf_in[4] == 0x00 && buf_in[5] == 0x00) {
    if(debug.verbose()) debug.print("Already unlocked?");
    return true;
  }
  // -- GENERATE KEY RESPONSE
  buf_out[BMSG] = 0x02; // send key msg value
  buf_out[4] = ~buf_in[4]; // bitwise NOT breaks the key ;]
  buf_out[5] = ~buf_in[5]; // bitwise NOT breaks the key ;]
  // -- ANTICIPATED REPLY
  reply[BMSG] = 0x02; // send key msg value
  reply[4] = 0xAA; // success code
  if(aldl_request_reply_specific(7,reply,6,75) == false) {
    debug.print("FLASH: Failed @ key submit stage");
    if(attempts >= 3) {
      flash_err.set_errstring("The ECM doesn't like our security code.  Power cycle and try again.");
      return false; // UP TO 2 TRIES
    }
    attempts++;
    sleep_ms(100);
    goto RETRY_KEY_SEED_REQUEST;
  }
  if(debug.verbose()) debug.print("Successful unlock routine!");
  return true;
}

bool datastream::upload_eside_development_routine() {
#ifdef FLASH_NO_ERR
  // this gets us into programming mode without caring about the response...
  aldl_request_reply_none(ESIDE_DEV_ROUTINE_A,sizeof(ESIDE_DEV_ROUTINE_A),250);
  aldl_request_reply_none(ESIDE_DEV_ROUTINE_B,sizeof(ESIDE_DEV_ROUTINE_B),250);
  aldl_request_reply_none(ESIDE_DEV_ROUTINE_C,sizeof(ESIDE_DEV_ROUTINE_C),250);
  aldl_request_reply_none(ESIDE_DEV_ROUTINE_D,sizeof(ESIDE_DEV_ROUTINE_D),250);
  serial_discard_all(250);
  return true;
#endif
  flash_err.reset();
  if(debug.verbose()) debug.print("Sending MEMORY CONTROL LOOP to e-side: ");
  if(aldl_request_reply_undefined(ESIDE_DEV_ROUTINE_A,sizeof(ESIDE_DEV_ROUTINE_A)) == 0) {
    if(buf_in[BMSG] != 0xAA) {  // kur4o's exception
      flash_err.set_errstring("Error during development OS stage A programming: ",buf_in[BMSG]);
      return false;
    }
  }
  serial_discard_all(50); // just in case, can't hurt.
  if(aldl_request_reply_undefined(ESIDE_DEV_ROUTINE_B,sizeof(ESIDE_DEV_ROUTINE_B)) == 0) {
    flash_err.set_errstring("Failed to execute upload stage B of the development OS!");
    return false;
  }
  if(buf_in[BMSG] != 0xAA) {
    flash_err.set_errstring("Error during development OS stage B programming: ",buf_in[BMSG]);
    return false;
  }
  if(aldl_request_reply_undefined(ESIDE_DEV_ROUTINE_C,sizeof(ESIDE_DEV_ROUTINE_C)) == 0 ) {
    flash_err.set_errstring("Failed to execute upload stage C of the development OS!");
    return false;
  }
  if(buf_in[BMSG] != 0xAA) {
    flash_err.set_errstring("Error during development OS stage C programming: ",buf_in[BMSG]);
    return false;
  }
  if(aldl_request_reply_undefined(ESIDE_DEV_ROUTINE_D,sizeof(ESIDE_DEV_ROUTINE_D)) == 0) {
    flash_err.set_errstring("Failed to execute upload stage D of the development OS!");
    return false;
  }
  if(buf_in[BMSG] != 0xAA) {
    flash_err.set_errstring("Error during development OS stage D programming: ",buf_in[BMSG]);
    return false;
  }
  return true;
}

bool datastream::upload_tside_development_routine() {
#ifdef FLASH_NO_ERR
  // this gets us into programming mode without caring about the response...
  aldl_request_reply_none(TSIDE_DEV_ROUTINE_A,sizeof(TSIDE_DEV_ROUTINE_A),250);
  aldl_request_reply_none(TSIDE_DEV_ROUTINE_B,sizeof(TSIDE_DEV_ROUTINE_B),250);
  aldl_request_reply_none(TSIDE_DEV_ROUTINE_C,sizeof(TSIDE_DEV_ROUTINE_C),250);
  aldl_request_reply_none(TSIDE_DEV_ROUTINE_D,sizeof(TSIDE_DEV_ROUTINE_D),250);
  serial_discard_all(250);
  return true;
#endif
  flash_err.reset();
  if(debug.verbose()) debug.print("FLASH: Sending MEMORY CONTROL LOOP to t-side: ");
  if(aldl_request_reply_undefined(TSIDE_DEV_ROUTINE_A,sizeof(TSIDE_DEV_ROUTINE_A)) == 0) {
    flash_err.set_errstring("Failed to execute upload stage A of the development OS!");
    return false;
  }
  if(buf_in[BMSG] != 0xAA) {
    flash_err.set_errstring("Error during development OS stage A programming: ",buf_in[BMSG]);
    return false;
  }
  if(aldl_request_reply_undefined(TSIDE_DEV_ROUTINE_B,sizeof(TSIDE_DEV_ROUTINE_B)) == 0) {
    flash_err.set_errstring("Failed to execute upload stage B of the development OS!");
    return false;
  }
  if(buf_in[BMSG] != 0xAA) {
    flash_err.set_errstring("Error during development OS stage B programming:  ",buf_in[BMSG]);
    return false;
  }
  if(aldl_request_reply_undefined(TSIDE_DEV_ROUTINE_C,sizeof(TSIDE_DEV_ROUTINE_C)) == 0 ) {
    flash_err.set_errstring("Failed to execute upload stage C of the development OS!");
    return false;
  }
  if(buf_in[BMSG] != 0xAA) {
    flash_err.set_errstring("Error during development OS stage C programming:  ",buf_in[BMSG]);
    return false;
  }
  if(aldl_request_reply_undefined(TSIDE_DEV_ROUTINE_D,sizeof(TSIDE_DEV_ROUTINE_D)) == 0) {
    flash_err.set_errstring("Failed to execute upload stage D of the development OS!");
    return false;
  }
  if(buf_in[BMSG] != 0xAA) {
    flash_err.set_errstring("Error during development OS stage D programming:  ",buf_in[BMSG]);
    return false;
  }
  return true;
}

bool datastream::upload_read_routine_tside() {
  flash_err.reset();
  if(debug.verbose()) debug.print("Sending READ ROUTINE to tside: ");
  if(aldl_request_reply_undefined(TSIDE_READ_ROUTINE,sizeof(TSIDE_READ_ROUTINE)) == 0) {
     flash_err.set_errstring("Failed to execute TSIDE_READ_ROUTINE");
     return false;
  }
  if(buf_in[BMSG] != 0xAA) {
    flash_err.set_errstring("TSIDE_READ_ROUTINE returned an error: ",buf_in[BMSG]);
    return false;
  }
  return true;
}

bool datastream::upload_read_routine_eside() {
  flash_err.reset();
  if(debug.verbose()) debug.print("Sending READ ROUTINE to eside: ");
  if(aldl_request_reply_undefined(ESIDE_READ_ROUTINE,sizeof(ESIDE_READ_ROUTINE)) == 0) {
     flash_err.set_errstring("Failed to execute ESIDE_READ_ROUTINE");
     return false;
  }
  if(buf_in[BMSG] != 0xAA) {
    flash_err.set_errstring("ESIDE_READ_ROUTINE returned an error: ",buf_in[BMSG]);
    return false;
  }
  return true;
}

bool datastream::remove_vpp_voltage() {
  flash_err.reset();
  if(debug.verbose()) debug.print("Disable VPP Voltage: ");
  // t-side routine A, disables VPP, etc.
  if(aldl_request_reply_undefined(TSIDE_VPP_REMOVE,sizeof(TSIDE_VPP_REMOVE),500) == 0) {
    flash_err.set_errstring("TSIDE_VPP_REMOVE routine upload error!");
    return false;
  }
  if(buf_in[BMSG] != 0xAA) {
    flash_err.set_errstring("TSIDE_VPP_REMOVE routine execution error: ",buf_in[BMSG]);
    return false;
  }
  return true;
}

bool datastream::reset_ecm_devmode() {
  flash_err.reset();
  if(debug.verbose()) debug.print("Reset e-side: ");
  bool result = true; // just scan for false results but continue....
  if(aldl_request_reply_undefined(ESIDE_RESET_MSG,sizeof(ESIDE_RESET_MSG),4000) == 0) {
    flash_err.set_errstring("ESIDE_RESET routine upload error!");
    result = false;
  }
  if(buf_in[BMSG] != 0xAA) {
    flash_err.set_errstring("ESIDE_RESET routine execution error!");
    result = false;
  }
  // final reset
  if(debug.verbose()) debug.print("Reset t-side: ");
  if(aldl_request_reply_undefined(TSIDE_RESET_MSG,sizeof(TSIDE_RESET_MSG),4000) == 0) {
    flash_err.set_errstring("TSIDE_RESET routine upload error!");
    result = false;
  }
  if(buf_in[BMSG] != 0xAA) {
    flash_err.set_errstring("TSIDE_RESET routine execution error! ",buf_in[BMSG]);
    result = false;
  }
  return result;
}

bool datastream::apply_vpp_voltage() {
  flash_err.reset();
  if(aldl_request_reply_undefined(TSIDE_VPP_APPLY,sizeof(TSIDE_VPP_APPLY)) == 0) {
    flash_err.set_errstring("Failed to upload program to apply VPP HI voltage");
    return false;
  }
  if(buf_in[BMSG] != 0xAA) {
    flash_err.set_errstring("Error trying to apply VPP voltage: ",buf_in[BMSG]);
    return false;
  }
  return true;
}

bool datastream::upload_erase_program(byte side) {
  flash_err.reset();
  serial_discard_all();
  if(side == 0xF4) {
    if(aldl_request_reply_undefined(TSIDE_PREP_A,sizeof(TSIDE_PREP_A)) == 0) {
      flash_err.set_errstring("T-side erase preperation stage failed to execute.");
      return false;
    }
    if(buf_in[BMSG] != 0xAA) {
      flash_err.set_errstring("T-side erase preperation stage returned an error: ",buf_in[BMSG]);
      return false;
    }
    if(aldl_request_reply_undefined(TSIDE_ERASE_A,sizeof(TSIDE_ERASE_A)) == 0) {
      flash_err.set_errstring("T-side erase program stage 1 failed to execute.");
      return false;
    }
    if(buf_in[BMSG] != 0xAA) {
      flash_err.set_errstring("T-side erase program stage 1 returned an error: ",buf_in[BMSG]);
      return false;
    }
    // this starts erase process:
    aldl_reqdup(TSIDE_ERASE_B,sizeof(TSIDE_ERASE_B)); // only send this command, reply is recieved in next function
    if(aldl_request_reply_none(sizeof(TSIDE_ERASE_B)) == false) {
      flash_err.set_errstring("Serial error sending second stage T-Side erase program");
      return false;
    }
    return true;
  } else if(side == 0xE4) {
    if(aldl_request_reply_undefined(ESIDE_PREP_A,sizeof(ESIDE_PREP_A)) == 0) {
      flash_err.set_errstring("E-side erase preperation stage failed to execute.");
      return false;
    }
    if(buf_in[BMSG] != 0xAA) {
      flash_err.set_errstring("E-side erase preperation stage returned an error: ",buf_in[BMSG]);
      return false;
    }
    if(aldl_request_reply_undefined(ESIDE_ERASE_A,sizeof(ESIDE_ERASE_A)) == 0) {
      flash_err.set_errstring("E-side erase program stage 1 failed to execute.");
      return false;
    }
    if(buf_in[BMSG] != 0xAA) {
      flash_err.set_errstring("T-side erase program stage 1 returned an error: ",buf_in[BMSG]);
      return false;
    }
    // this starts erase process:
    aldl_reqdup(ESIDE_ERASE_B,sizeof(ESIDE_ERASE_B)); // only send this command, reply is recieved in next function
    if(aldl_request_reply_none(sizeof(ESIDE_ERASE_B)) == false) {
      flash_err.set_errstring("Serial error sending second stage E-Side erase program");
      return false;
    }
    return true;
  }
  return false;
}

byte datastream::wait_for_erase_complete() {
  flash_err.reset();
  sleep_ms(50);
  if(aldl_read_reply(10000) == 0) return 0x00;
  return buf_in[BMSG];
}

bool datastream::upload_flash_program(byte side) {
  flash_err.reset();
  if(side == 0xF4) {
    if(aldl_request_reply_undefined(TSIDE_WRITE_ROUTINE,sizeof(TSIDE_WRITE_ROUTINE)) == 0) {
      flash_err.set_errstring("T-Side write program upload failure.");
      return false;
    }
    if(buf_in[BMSG] != 0xAA) {
      flash_err.set_errstring("T-Side write program error code: ",buf_in[BMSG]);
      return false;
    }
    return true;
  } else if(side == 0xE4) {
    if(aldl_request_reply_undefined(ESIDE_WRITE_ROUTINE,sizeof(ESIDE_WRITE_ROUTINE)) == 0) {
      flash_err.set_errstring("E-Side write program upload failure.");
      return false;
    }
    if(buf_in[BMSG] != 0xAA) {
      flash_err.set_errstring("E-Side write program error code: ",buf_in[BMSG]);
      return false;
    }
    return true;
  }
  return false; // invalid device
}

bool datastream::write_flash(byte side, bin_file bin, quint16 offset, byte length, int postdelay) {
  if(length == 0) return true; // no data
  flash_err.reset();
  // [8],[9] = addr [10]=size [11]=data start
  //[dev] [len] 06 03 00 BD 02 1a (ADDR) (ADDR) (NUMBER OF BYTES) [ 128 BYTES OF DATA ] CKSUM
  // configure packet header
  aldl_reqdup(WRITE_PACKET_PRIMER,sizeof(WRITE_PACKET_PRIMER)); // pre-configure most of the packet
  buf_out[BDEVICE] = side; // set side
  // set offset
  buf_out[8] = (offset >> 8) & 0x00FF; /* memory location high byte */
  buf_out[9] = offset & 0x00FF; /* memory location low byte */
  // set length
  buf_out[10] = length; // +12=for total length incl checksum
  // get ptr to data segment in bin
  byte *data;
  if(side == 0xF4) { // tside
    data = bin.tside + offset;
  } else { // eside
    data = bin.eside + offset;
  }
  // copy data from bin to buffer
  for(int x=0; x<length;x++) {
     buf_out[11 + x] = data[x];
  }
  // send request
  if(aldl_request_reply_undefined(length + 12,1000,postdelay) == false) {
    flash_err.set_errstring("Write packet upload error.");
    return false;
  }
  if(buf_in[BMSG] != 0xAA) {
    flash_err.set_errstring("Write packet execution error code: ",buf_in[BMSG]);
    return false;
  }
  return true;
}

bool datastream::get_post_status() {
  sleep_ms(2000);
  flash_err.reset();
  // the ecm is now awake... ybody shouldn't have traffic here.
  silence_modules();
  if(m8_silence_request() == false) {
    flash_err.set_errstring("Can't communicate with T-Side!!");
    return false;
  }
  // wait for err code generation
  for(int x=0;x<3;x++) {
    if(m8_silence_request() == false) {
      flash_err.set_errstring("Can't communicate with T-Side!!");
      return false;
    }
  }
  sleep_ms(200);
  set_device(PRIMARY_ECM);
  set_mode(0x01); // mode
  set_msg(0x00); // message
  if(aldl_request_reply_undefined(5,1500) == false) {
    flash_err.set_errstring("Can't get ECM status!!");
    return false;
  }

  if(getbit(1,buf_in[10]) == true) {
    flash_err.set_errstring("EEPROM Checksum Failure!!  Possible write error.\nTry flashing again...");
    return false;
  }
  return true;
}
