﻿#include <QApplication>
#include "config.h"
#include "datastream.h"
#include "bin_file.h"
#include "common.h"

bool datastream::check_if_ok_to_flash() {
  return true; // all passed
}

void datastream::flash_message(QString msg) {
  emit new_flash_message(msg);
}

void datastream::flash_error_message(QString msg) {
  emit new_flash_error(msg);
}

void datastream::flash_error() {
  emit flash_error_occurred();
}

void datastream::flash_divider(QString msg) {
  emit new_flash_divider(msg);
}

void datastream::flash_cancel_enable(bool value) {
  emit enable_flash_exit(value);
}

void datastream::flash_speed(float value) {
  emit update_flash_speed(value);
}

void datastream::flash_tside_progress(int value) {
  if(value<0x2000) return;
  emit update_tside_progress(value);
}

void datastream::flash_eside_progress(int value) {
  if(value<0x2000) return;
  emit update_eside_progress(value);
}

void datastream::flash_close_window() {
  emit end_flashprogress();
}

void datastream::write_flash_procedure() {
  // THIS IS A MEGAFUNCTION FROM HELL!!!!! ... but it works well.
  // It's kept in one function so workflow can use lots of GOTOs and other
  // bad programming practices.  Modify with caution!!

  // for iterating through e-side and t-side in loops...
  byte side[2] = { PRIMARY_ECM, SECONDARY_ECM };

  // ecm rst canary
  bool ecm_was_reset = true;
  bool safe_mode = false;
  bool vpp_applied = false;

  // figure which sides to write
  bool enable[2] = { false, false };
  if(control->flash_tside.get() == true) enable[0] = true;
  if(control->flash_eside.get() == true) enable[1] = true;
  if(enable[0] == false && enable[1] == false) { // FIXME this could be done with button disables...
    emit(display_critical_error("Must select either T-Side or E-Side to flash!"));
    return;
  }

  // timer
  QElapsedTimer block_write_time;
  QElapsedTimer total_write_time;

  // skip unused regions?
  bool skip_unused_regions = control->skip_unused_regions.get();

  // get timing
  int write_post_delay = FLASH_WRITE_POST_DELAY_LONG;
  if(config.get_flash_low_latency() == true) {
    write_post_delay = FLASH_WRITE_POST_DELAY_SHORT;
  }

  // get block size settings
  int write_max = DEFAULT_WRITE_BLOCK_SIZE;

  // work on a copy of the bin, so we don't mangle the one in memory
  bin_file bin;
  control->write_bin.copy_to(&bin);

  // signal for dialog to open and wait a sec
  // FIXME this should actually wait for a window open signal or something
  emit(start_flashprogress());
  flash_divider("INFORMATION");
  flash_message("Vin# " + log->info.vin_number.get());
  flash_message(bin.ecm_info_string());
  flash_message(QString("Write block size %1 bytes").arg(write_max));
  flash_message(QString("Write post-delay %1 ms").arg(write_post_delay));
  if(skip_unused_regions == true) flash_message(QString("Skip 0xFF Enabled"));

  flash_divider("PREPARING TO WRITE");

  flash_cancel_enable(true);
  if(check_if_ok_to_flash() == false) {
    flash_error_message("Some conditions exist that would make flash i/o unsafe.");
    flash_error_message("See error messages above.");
    goto FLASH_WRITE_CLEANUP;
  }
  if(control->flash_exit.get() == true) return;

  // simplifier
  if(skip_unused_regions == true) {
    flash_message(QString("Erased %1 unused bytes with 0xFF").arg(bin.ff_unused_regions()));
  }

  // patch
  if(control->enable_patches.get() == true) {
    if(log->info.patch_version.get() < CURRENT_PATCH_VERSION) { // need to flash both sides
      flash_error_message("Forcing two sided write due to patch upgrade!");
      enable[0] = true;
      enable[1] = true;
    }
    if(bin.apply_patches() == false) {
      flash_error_message("Patch application failed for some reason!");
      flash_error_message("Patches can only be applied to a virgin based bin,");
      flash_error_message("If you read a patched bin and try to flash it again,");
      flash_error_message("it'll fail...");
      goto FLASH_WRITE_CLEANUP;
    } else {
      flash_message(QString("Patched up to version %1").arg(CURRENT_PATCH_VERSION));
    }
  }

  // update checksum
  bin.reset_checksums();

  flash_divider("PREPARE FOR PROGRAMMING");

  // check again if connected
  if(control->is_connected() == false) {
    flash_error_message("Connection lost, please try again.");
    goto FLASH_WRITE_CLEANUP;
  }

  // lets check now if we've been connected long enough to ensure rst timeout
  if(control->time_connected.elapsed() < FLASH_MIN_RST_TIME) {
    flash_message("ECM RESET TIMER MAY STILL BE ACTIVE, PLEASE WAIT...");
    while(control->time_connected.elapsed() < FLASH_MIN_RST_TIME) {
      tickle_ecm();
      sleep_ms(500);
    }
  }

  // break security
  flash_message("Disabling programming lockouts (Mode 13)");
  for(int x=0;x<=1;x++) {
    if(unlock_security(side[x]) == false) {
      flash_error_message(flash_err.get_errstring());
      flash_message("This might just be a communication glitch.  Try key on/off and go again.");
      goto FLASH_WRITE_CLEANUP;
    }
  }

  flash_message("Uploading T-Side development routine...");
  if(enter_mode5(0xF4) == false) { // enter programming mode
    flash_error_message(flash_err.get_errstring());
    flash_message("Access can be denied for a number of reasons, including low PCM voltage, or your car is running/moving.");
    goto FLASH_WRITE_CLEANUP;
  }
  if(control->flash_exit.get() == true) goto FLASH_WRITE_CLEANUP;
  // from this point on, a reset is required !!
  ecm_was_reset = false;
  if(upload_tside_development_routine() == false) {
    flash_error_message(flash_err.get_errstring());
    flash_error();
    goto FLASH_WRITE_CLEANUP;  // failure here means no chance of reset
  }
  if(control->flash_exit.get() == true) goto FLASH_WRITE_RESET;

  // eside mode5 and dev
  flash_message("Uploading E-Side development routine...");
  if(enter_mode5(0xE4) == false) { // enter programming mode
    flash_error_message(flash_err.get_errstring());
    flash_message("Access can be denied for a number of reasons, including low PCM voltage, or your car is running/moving.");
    flash_error();
    goto FLASH_WRITE_CLEANUP; // must reset since t-side already loaded
  }
  if(control->flash_exit.get() == true) goto FLASH_WRITE_RESET;
  if(upload_eside_development_routine() == false) {
    flash_error_message(flash_err.get_errstring());
    flash_error();
    goto FLASH_WRITE_CLEANUP;  // failure here means no chance of reset
  }
  if(control->flash_exit.get() == true) goto FLASH_WRITE_RESET;

  flash_message("Both ECMs are now executing code from RAM only.");

  // apply vpp voltage
  flash_message("Applying VPP (EEPROM Programming) voltage");
  if(apply_vpp_voltage() == false) {
    flash_error_message(flash_err.get_errstring());
    flash_error();
    goto FLASH_WRITE_CLEANUP;  // failure here means no chance of reset
  } else {
    vpp_applied = true;
  }
  if(control->flash_exit.get() == true) goto FLASH_WRITE_RESET;

  flash_cancel_enable(false);

  // iterate through tside and eside
  for(int current_side=0;current_side<=1;current_side++) {

    int flash_restart_attempts = 0; // reset for next side
    flash_speed(0); // set to zero bps for now

    // get ptr to write data itself
    byte *write_data = nullptr;
    if(side[current_side] == 0XF4) write_data = (byte *)(bin.tside);
    if(side[current_side] == 0XE4) write_data = (byte *)(bin.eside);
    if(write_data == nullptr) continue; // invalid side?  should never happen

    // log header
    if(side[current_side] == PRIMARY_ECM) {
      if(enable[current_side] == false) {
        flash_divider("SKIPPING T-SIDE");
        continue; // skip non-enabled side
      } else {
        flash_divider("ERASING T-SIDE");
      }
    } else {
      if(enable[current_side] == false) {
        flash_divider("SKIPPING E-SIDE");
        continue; // skip non-enabled side
      } else {
       flash_divider("ERASING E-SIDE");
      }
    }

    FLASH_WRITE_ERASE_RESTART:
    if(flash_restart_attempts > FLASH_WRITE_MAX_RESTARTS) {
      flash_error_message("Recovery mode is giving up :(");
      flash_error_message("Something has gone wrong flashing your ECM.");
      flash_error_message("There's a chance your ECM is still alive, though.  We'll try to reset now...");
      goto FLASH_WRITE_RESET;
    }

    // upload erase routine
    flash_message("Uploading ERASE program:");
    if(upload_erase_program(side[current_side]) == false) {
      flash_error_message(flash_err.get_errstring());
      flash_restart_attempts++;
      QThread::msleep(8000); // delay
      flash_error();
      goto FLASH_WRITE_ERASE_RESTART;
    }

    // wait for erase complete
    int erase_stage = 0;
    int erase_fail = 0;
    forever {
      if(erase_fail > 4) { // too many fails
        flash_error_message("Attempting recovery, restarting erase procedure...");
        QThread::msleep(5000); // delay
        flash_restart_attempts++;
        flash_error();
        goto FLASH_WRITE_ERASE_RESTART;
      }
      byte erase_response = wait_for_erase_complete();
      if(erase_response == 0x00) {
        flash_error_message("Erase timeout!");
        erase_fail++;
        continue;
      }
      if(erase_response == 0x55) {
        erase_stage++;
        erase_fail = 0;
        flash_message(QString("  Erase progress %1%").arg(erase_stage * 25));
        continue;
      }
      if(erase_response == 0xAA) {
        erase_stage++;
        erase_fail = 0;
        flash_message(QString("  Erase Complete"));
        break;
      }
      // odd response
      flash_error_message("  Bad Response!");
      erase_fail++;
    }

    QThread::msleep(75); // force post-erase delay

    // log header
    if(side[current_side] == PRIMARY_ECM) {
      flash_divider("PROGRAMMING T-SIDE");
    } else {
      flash_divider("PROGRAMMING E-SIDE");
    }

    // upload flash program
    flash_message("Uploading FLASH WRITE program...");
    // we run this in a loop, just as a last ditch effort if it flunks, as failure here will brick ecm.
    // this is a very dangerous place to die as the eeprom is certainly erased already.
    while(upload_flash_program(side[current_side]) == false) {
      flash_cancel_enable(true);
      if(control->flash_exit.get() == true) goto FLASH_WRITE_CLEANUP;
      flash_error_message(flash_err.get_errstring());
      flash_message("This is BAD NEWS!  Attempting recovery...");
      flash_message("This runs in an infinite loop... you can click cancel if you like.");
      QThread::msleep(5000); // delay
      flash_error();
      flash_divider("RECOVERY");
    }
    flash_cancel_enable(false);

    flash_message("Writing EEPROM ... See progress bars");

    // configuration
    int write_offset = 0x2000; // starting offset
    int unused_skipped_total = 0;
    int unused_skipped_this_iteration = 0;

    flash_speed(0);

    total_write_time.start(); // start timer

    // loop through bin
    while(write_offset < 0x10000) {

      // 0xFF stage A (we are starting with 0xFF, skipping it is free..)
      unused_skipped_this_iteration = 0;
      while(skip_unused_regions == true &&
            write_offset < 0xFFFA &&
            write_data[write_offset] == 0xFF) {
        write_offset++;
        unused_skipped_total++;
        unused_skipped_this_iteration++;
      }

      // determine base block size
      int write_size = clamp(0xFFFF - write_offset + 1,1,write_max);

      // reduce write end size.  on the next iteration, this will be skipped in stage A too.
      while(skip_unused_regions == true && write_offset < 0xE000 &&
            write_data[write_offset+write_size - 1] == 0xFF &&
            write_size > 2) {
        write_size--;
      }

      int write_retry_attempts = 0; // rst

      // write data, loop if fail.
      block_write_time.start();
      while(write_flash(side[current_side],bin,write_offset,write_size,write_post_delay) == false) {
        QThread::msleep(1000); // delay 1 second on a bad write.
        flash_error();
        flash_error_message(flash_err.get_errstring());
        write_retry_attempts++;
        // set failsafes even after a single error
        if(safe_mode == false) {
          flash_error_message("ENABLE SAFE MODE: Disable large block size, and fast write!");
          write_post_delay = FLASH_WRITE_POST_DELAY_SAFE; // slow down post delay ?
          write_max = 0x80; // use smaller write size?
          safe_mode = true;
        }
        if(write_retry_attempts > FLASH_WRITE_MAX_RETRIES) { // we're stuck gettng errors.
          flash_error_message("Too many sequential write errors!!!");
          flash_divider("RECOVERY");
          flash_error_message("Entering flash recovery mode (Re-erase and start over...)");
          flash_restart_attempts++;
          goto FLASH_WRITE_ERASE_RESTART;
        }
        flash_message("Re-sending packet...");
      }

      // update progress sliders
      if(side[current_side] == 0xF4) {
        flash_tside_progress(write_offset + write_size);
      } else {
        flash_eside_progress(write_offset + write_size);
      }
      flash_speed(((float)(write_size +  unused_skipped_this_iteration)
                                / (float)block_write_time.elapsed()) * 1000 * 8);

      // update offset
      write_offset += write_size;
    }
    // display skip percentage
    flash_message(QString("Skipped " +
                                QString::number(
                                  ((float)unused_skipped_total / (float)0xE000) * 100.00,'f',1)
                                ) + "% unused regions");
    flash_message(QString("Write Time: " + QString::number(total_write_time.elapsed() / 1000,'f',0) + " Seconds"));
  }

  flash_divider("EXIT PROGRAMMING MODE");
  flash_speed(0);

  FLASH_WRITE_RESET:
  if(vpp_applied == true) {
    flash_message("Removing VPP Voltage");
    if(remove_vpp_voltage() == false) {
      flash_error();
      flash_error_message(flash_err.get_errstring());
    }
    vpp_applied = false;
  }

  flash_message("Resetting ECM and exiting development routine");
  if(reset_ecm_devmode() == false) {
    flash_error();
    flash_error_message(flash_err.get_errstring());
  } else {
    ecm_was_reset = true;
  }

  // success
  flash_message("Flash WRITE Complete!");

  FLASH_WRITE_CLEANUP:
  if(ecm_was_reset == false) {
    flash_error_message("*************************");
    flash_error_message("Unfortunately, the development routine is now stuck in ram...");
    flash_error_message("This means you need to disconnect power from your ecm for a bit");
    flash_error_message("to reset it manually before it'll function again.");
    flash_error_message("*************************");
  } else {
    flash_message("Testing ECM Functionality...");
    if(get_post_status() == false) {
      flash_error();
      flash_error_message(flash_err.get_errstring());
      flash_message("This might not actually be a bad thing.  Perhaps you should");
      flash_message("try cycling the key and starting the car anyway.");
    } else {
      flash_message("Everything looks good!");
      emit flash_write_successful();
    }
  }

  // exit
  flash_divider("FINISHED");
  flash_message("Press EXIT.  Will disconnect automatically.");
  flash_cancel_enable(true);
  while(control->flash_exit.get() == false) msleep(20);
  flash_close_window();
  return;
}

void datastream::read_flash_procedure() {
  bin_file bin;
  byte side[2] = { PRIMARY_ECM, SECONDARY_ECM };

  bool ecm_was_reset = false;
  bool read_success = false;
  bool ram_notify = false;

  // determine read floor
  int read_floor = 0x0D80;  // below secondary eeprom
  if(control->dump_ram.get() == true) read_floor = 0x0000;

  // open window
  emit(start_flashprogress());
  msleep(200);

  // timer
  QElapsedTimer block_read_time;
  QElapsedTimer total_read_time;
  total_read_time.start();

  flash_divider("EXECUTING FLASH READ ROUTINE");
  if(control->dump_ram.get() == true) flash_message(QString("Ram dump enabled"));

  flash_cancel_enable(true);

  // warn on version>0
  if(log->info.patch_version.get() > 0x00) {
    flash_divider("PATCH WARNING");
    flash_error_message("WARNING!  The bin on this ECM has been patched with EEHack.");
    flash_error_message("The bin you are now reading may yeild odd results in bin editing software.");
    flash_error_message("If uploading this bin again, it may refuse to re-patch.");
  }

  // break security
  flash_message("Disabling programming lockouts (Mode 13)");
  for(int x=0;x<=1;x++) {
    if(unlock_security(side[x]) == false) {
      flash_error_message(flash_err.get_errstring());
      flash_message("This might just be a communication glitch.  Try key on/off and go again.");
      goto FLASH_READ_CLEANUP_NO_RST;
    }
  }

  flash_message("Uploading T-Side development routine...");
  if(enter_mode5(0xF4) == false) { // enter programming mode
    flash_error_message(flash_err.get_errstring());
    flash_message("Access can be denied for a number of reasons, including low PCM voltage, or your car is running/moving.");
    goto FLASH_READ_CLEANUP_NO_RST;
  }
  if(control->flash_exit.get() == true) goto FLASH_READ_CLEANUP;
  // from this point on, a reset is required !!
  ecm_was_reset = false;
  if(upload_tside_development_routine() == false) {
    flash_error_message(flash_err.get_errstring());
    flash_error();
    goto FLASH_READ_CLEANUP;  // failure here means no chance of reset
  }
  if(control->flash_exit.get() == true) goto FLASH_READ_CLEANUP;

  // eside mode5 and dev
  flash_message("Uploading E-Side development routine...");
  if(enter_mode5(0xE4) == false) { // enter programming mode
    flash_error_message(flash_err.get_errstring());
    flash_message("Access can be denied for a number of reasons, including low PCM voltage, or your car is running/moving.");
    flash_error();
    goto FLASH_READ_CLEANUP; // must reset since t-side already loaded
  }
  if(control->flash_exit.get() == true) goto FLASH_READ_CLEANUP;
  if(upload_eside_development_routine() == false) {
    flash_error_message(flash_err.get_errstring());
    flash_error();
    goto FLASH_READ_CLEANUP;  // failure here means no chance of reset
  }
  if(control->flash_exit.get() == true) goto FLASH_READ_CLEANUP;

  flash_message("Both ECMs are now executing code from RAM only.");

  //program
  flash_message("Executing Read Program...");
  if(upload_read_routine_eside() == false) {
    flash_error_message(flash_err.get_errstring());
    flash_error();
    goto FLASH_READ_CLEANUP;
  }
  if(upload_read_routine_tside() == false) {
    flash_error_message(flash_err.get_errstring());
    flash_error();
    goto FLASH_READ_CLEANUP;
  }

  //read
  flash_cancel_enable(true); // ok to cancel during read.
  flash_divider("Reading EEPROM");
  flash_message("<-- See Progress");

  // read main iterator
  for(int x=0xFF80;x>=read_floor;x-=0x80) {
    block_read_time.start();
    // progress
    flash_tside_progress(x);
    flash_eside_progress(x);
    if(x < 0x2000 && ram_notify == false) {
      flash_divider("Reading Extended Memory (RAM/Secondary EEPROM)");
      ram_notify = true;
    }

    if(read_floor != 0x0000) { // we aren't dumping ram
      if(x < 0x0D00) continue; // below secondary eeprom
      if(x > 0x1080 && x < 0x1F80) continue; // above secondary eeprom but below main prom
    }

    if(control->flash_exit.get() == true) {
      flash_message("Cancelled FLASH READ.");
      break;
    }
    if(read_flash_section(PRIMARY_ECM,bin.tside,x) == false) {
      flash_error_message(flash_err.get_errstring());
      flash_error();
      goto FLASH_READ_CLEANUP;
    }

    if(read_flash_section(SECONDARY_ECM,bin.eside,x) == false) {
      flash_error_message(flash_err.get_errstring());
      flash_error();
      goto FLASH_READ_CLEANUP;
    }
    flash_speed(((float)(0x80 * 2 / (float)block_read_time.elapsed()) * 1000 * 8));
  }

  // success
  flash_message("Flash READ Complete!");
  if(read_floor != 0x0000) bin.zero_ram(); // if not reading ram, clear ram region
  flash_message(QString("Read Time Total: " + QString::number(total_read_time.elapsed() / 1000,'f',0) + " Seconds"));

  if(bin.test_checksum() == false) {
    flash_error_message("Checksum FAIL in downloaded bin!  Bin might be corrupt.");
    if(config.get("ALLOW_BAD_CKSUM",false) == true) {
      flash_error_message("Your settings allow saving bins with bad checksums.  We'll pretend this is okay..");
      read_success = true;
    }
  } else {
    flash_message("Checksum OK!");
    flash_message("Looks like everything worked out alright.  You will be prompted to save the bin after pressing 'exit'.");
    bin.copy_to(&(control->read_bin));
    read_success = true;
  }

  FLASH_READ_CLEANUP:
  flash_message("Reset ECM and exit development mode...");
  if(reset_ecm_devmode() == false) {
    flash_error_message(flash_err.get_errstring());
  } else {
    ecm_was_reset = true;
  }

  if(ecm_was_reset == false) {
    flash_error();
    flash_message("*************************");
    flash_message("Unfortunately, the development routine is now stuck in ram :(");
    flash_message("This means you need to disconnect power from your ecm for a bit");
    flash_message("to reset it before it'll function normally; pulling the PCM BAT");
    flash_message("fuse would do the trick!!");
    flash_message("*************************");
  }

  FLASH_READ_CLEANUP_NO_RST:
  flash_divider("FINISH");
  flash_message("Press EXIT to leave flash read routine.");
  flash_cancel_enable(true);
  while(control->flash_exit.get() == false) sleep_ms(20);
  flash_close_window();
  if(read_success == true) emit flash_read_successful();
  return;
}

void datastream::flash_stability_test() {
  bin_file bin;
  byte side[2] = { PRIMARY_ECM, SECONDARY_ECM };
  bool ecm_was_reset = true;

  // open window
  emit(start_flashprogress());
  msleep(200);

  // timer
  QElapsedTimer block_read_time;
  QElapsedTimer total_read_time;
  total_read_time.start();

  flash_divider("EXECUTING PROGRAMMING STABILITY TEST");

  flash_cancel_enable(true);

  // break security
  flash_message("Disabling programming lockouts (Mode 13)");
  for(int x=0;x<=1;x++) {
    if(unlock_security(side[x]) == false) {
      flash_error_message(flash_err.get_errstring());
      flash_message("This might just be a communication glitch.  Try key on/off and go again.");
      goto FLASH_STAB_CLEANUP_NO_RST;
    }
  }

  flash_message("Uploading T-Side development routine...");
  if(enter_mode5(0xF4) == false) { // enter programming mode
    flash_error_message(flash_err.get_errstring());
    flash_message("Access can be denied for a number of reasons, including low PCM voltage, or your car is running/moving.");
    goto FLASH_STAB_CLEANUP_NO_RST;
  }
  if(control->flash_exit.get() == true) goto FLASH_STAB_CLEANUP;
  if(upload_tside_development_routine() == false) {
    flash_error_message(flash_err.get_errstring());
    flash_error();
    goto FLASH_STAB_CLEANUP;  // failure here means no chance of reset
  }
  if(control->flash_exit.get() == true) goto FLASH_STAB_CLEANUP;
  ecm_was_reset = false;
  // eside mode5 and dev
  flash_message("Uploading E-Side development routine...");
  if(enter_mode5(0xE4) == false) { // enter programming mode
    flash_error_message(flash_err.get_errstring());
    flash_message("Access can be denied for a number of reasons, including low PCM voltage, or your car is running/moving.");
    flash_error();
    goto FLASH_STAB_CLEANUP; // must reset since t-side already loaded
  }
  if(control->flash_exit.get() == true) goto FLASH_STAB_CLEANUP;
  if(upload_eside_development_routine() == false) {
    flash_error_message(flash_err.get_errstring());
    flash_error();
    goto FLASH_STAB_CLEANUP;  // failure here means no chance of reset
  }
  if(control->flash_exit.get() == true) goto FLASH_STAB_CLEANUP;

  flash_message("Both ECMs are now executing code from RAM only.");

  flash_cancel_enable(true); // ok to cancel during read.

  static int n_iterations = 15;
  for(int iterations=1;iterations<=n_iterations;iterations++) {
    if(control->flash_exit.get() == true) {
      flash_message("Cancelling stability test...");
      break;
    }

    flash_message("Stability test iteration " + QString::number(iterations) + "/" + QString::number(n_iterations));

    if(upload_read_routine_eside() == false) {
      flash_error_message(flash_err.get_errstring());
      flash_error();
      goto FLASH_STAB_CLEANUP;
    }
    if(upload_read_routine_tside() == false) {
      flash_error_message(flash_err.get_errstring());
      flash_error();
      goto FLASH_STAB_CLEANUP;
    }

    for(int x=0x42ff;x>=0x4000;x-=0x80) {
      if(control->flash_exit.get() == true) break;
      if(read_flash_section(PRIMARY_ECM,bin.tside,x) == false) {
        flash_error_message(flash_err.get_errstring());
        flash_error();
        goto FLASH_STAB_CLEANUP;
      }
      msleep(qrand() % 8); // timing wobble
      if(read_flash_section(SECONDARY_ECM,bin.eside,x) == false) {
        flash_error_message(flash_err.get_errstring());
        flash_error();
        goto FLASH_STAB_CLEANUP;
      }
    }

    if(remove_vpp_voltage() == false) {
      flash_error();
      flash_error_message(flash_err.get_errstring());
    }
  }

  // success
  flash_message("Stability test complete.  If there are no errors above, it's probably safe to flash your ECM.");

  FLASH_STAB_CLEANUP:
  flash_message("Reset ECM and exit development mode...");
  if(reset_ecm_devmode() == false) {
    flash_error_message(flash_err.get_errstring());
  } else {
    ecm_was_reset = true;
  }

  if(ecm_was_reset == false) {
    flash_error();
    flash_message("*************************");
    flash_message("Unfortunately, the development routine is now stuck in ram :(");
    flash_message("This means you need to disconnect power from your ecm for a bit");
    flash_message("to reset it before it'll function normally; pulling the PCM BAT");
    flash_message("fuse would do the trick!!");
    flash_message("*************************");
  }

  FLASH_STAB_CLEANUP_NO_RST:
  flash_divider("FINISH");
  flash_message("Press EXIT to leave flash stability test.");
  flash_cancel_enable(true);
  while(control->flash_exit.get() == false) sleep_ms(20);
  flash_close_window();
  return;
}
